from abc import ABC, abstractmethod
from typing import Union, Tuple
import logging
import scipy
import numpy as np
from scipy.sparse import eye, csr_matrix
from scipy.sparse.linalg import LinearOperator

from .util import StopWatch
from .operators import MatrixOperator, InverseWeightOperator
from .regularization import RegularizationRule, FixedRankSpectralShiftRegularizationRule

from .weighted_least_squares import (
    WeightedLeastSquaresSolver,
    ScipyCgWeightedLeastSquaresSolver,
)

log = logging.getLogger(__name__)


class StoppingCriteria(ABC):
    """
    Abstract base class for iteration stopping criteria
    """
    @abstractmethod
    def satisfied(self, previous_iterate: np.ndarray, current_iterate: np.ndarray) -> bool:
        """
        Specifies if criteria is satisfied.
        :param current_iterate:
        :param previous_iterate:
        :return:
        """
        pass


class ResidualNormStoppingCriteria(StoppingCriteria):
    def __init__(self, tol=1e-9):
        """
        Stopping criteria based on the relative residual between two iterations is less than a given tolerance
        :param tol: tolerance for residual
        """
        self._tol = tol
        self._log = log.getChild(self.__class__.__name__)

    @property
    def tol(self):
        return self._tol

    def satisfied(self, previous_iterate: np.ndarray, current_iterate: np.ndarray):
        """
        Returns true if

        .. math::

            \\frac{\\|X_{\\text{prev}} - X_{\\text{curr}}\\||_{F}}{\\|X_{\\text{prev}}\\|_{F}} < tol.

        :param previous_iterate: :math:`X_{\\text{prev}`
        :param current_iterate: :math:`X_{\\text{curr}`
        :return:
        """
        residual = np.linalg.norm(previous_iterate - current_iterate) / np.linalg.norm(previous_iterate)
        self._log.info(f"Current {residual=:.6f}")
        return (
                residual
                < self.tol
        )


class Problem:
    def __init__(
        self,
        measurement_operator: Union[LinearOperator, np.ndarray, csr_matrix, MatrixOperator],
        data: np.array,
        indexing_order: MatrixOperator.IndexingOrder = None,
        input_shape: Tuple[int, int] = None,
    ):
        """
        Description for a rank minimization problem

        .. math::

            \\min_{x \\in \\mathbb{C}^{d_1 \\times d_2}} \\operatorname{rank}(x), \\text{s.t.} \\Phi(x) = y,


        :param measurement_operator:
        :param data:
        :param indexing_order:
        :param input_shape:
        """

        if not isinstance(measurement_operator, MatrixOperator):
            if indexing_order is None or input_shape is None:
                raise ValueError(
                    f"In case provided operator is not wrapped as MatrixOperator, "
                    f"provide indexing order and input shape"
                )
            measurement_operator = MatrixOperator(
                measurement_operator, input_shape, data.shape, indexing_order
            )

        self.data = data
        self.measurement_operator = measurement_operator
        self._log = log.getChild(self.__class__.__name__)

    def solve(
        self,
        schatten_p_parameter,
        max_iter=1000,
        rank_estimate: int = None,
        regularization_rule: RegularizationRule = None,
        weighted_least_squares_solver: WeightedLeastSquaresSolver = ScipyCgWeightedLeastSquaresSolver(
            tol=1e-5
        ),
        stopping_criteria: StoppingCriteria = ResidualNormStoppingCriteria(tol=1e-4),
    ):
        """

        :param schatten_p_parameter:
        :param max_iter:
        :param rank_estimate:
        :param regularization_rule:
        :param weighted_least_squares_solver:
        :param stopping_criteria:
        :return:
        """

        if regularization_rule is None:
            if rank_estimate is None:
                raise ValueError(
                    f"Provide either a rank estimate or a regularization rule."
                )
            regularization_rule = FixedRankSpectralShiftRegularizationRule(
                rank_estimate
            )

        def _initialize_inverse_weight_matrix_operator():
            """
            Initialize inverse weight matrix operator with identity matrix
            :return: MatrixOperator
            """
            flattened_input_length = np.product(self.measurement_operator.input_shape)
            order = self.measurement_operator.order
            input_shape = self.measurement_operator.input_shape
            return MatrixOperator(
                scipy.sparse.linalg.aslinearoperator(eye(flattened_input_length)),
                input_shape,
                input_shape,
                representing_matrix=eye(flattened_input_length),
                order=order,
            )

        inverse_weight_matrix_operator = _initialize_inverse_weight_matrix_operator()
        iteration = 0
        old_result = np.random.randn(*self.measurement_operator.input_shape)
        result = None
        while iteration < max_iter:
            with StopWatch() as iteration_stop_watch:
                with StopWatch() as solve_stop_watch:
                    result = weighted_least_squares_solver.solve(
                        self.measurement_operator, inverse_weight_matrix_operator, self.data
                    )
                self._log.debug(f"Solving weighted least squares took {solve_stop_watch.duration} seconds")

                with StopWatch() as weight_update_stop_watch:
                    if stopping_criteria.satisfied(old_result, result):
                        return result
                    old_result = result
                    (
                        left_inverse_weight,
                        right_inverse_weight,
                    ) = regularization_rule.compute_regularized_inverse_weights(
                        result, schatten_p_parameter
                    )
                    inverse_weight_matrix_operator = InverseWeightOperator(
                        left_inverse_weight,
                        right_inverse_weight,
                        self.measurement_operator.order,
                    )
                self._log.debug(f"Updating weight matrices took {weight_update_stop_watch.duration} seconds")
            self._log.info(f"Finished {iteration=} in {iteration_stop_watch.duration:.2f} seconds")
            iteration += 1
        return result



from abc import ABC, abstractmethod
import numpy as np
from scipy.sparse import diags
from hmirls.svd import ScipySVDEngine, SVDEngine


class RegularizationRule(ABC):

    @abstractmethod
    def compute_regularized_inverse_weights(
        self, x: np.ndarray, schatten_p_parameter: np.float64
    ):
        """
        Abstraction of two regularization functions

        .. math::

            R_1 : \\mathbb{C}^{d_1 \\times d_1} \\rightarrow \\mathbb{C}^{d_1 \\times d_1},
            R_2 : \\mathbb{C}^{d_2 \\times d_2} \\rightarrow \\mathbb{C}^{d_2 \\times d_2}

        representing a regularization rule for :math:`xx^{\\star}`,resp. :math:`x^{\\star}x`

        :param x: array of shape :math:`(d_1, d_2)`
        :param schatten_p_parameter:
        :return: :math:`R_1(xx^{\\star})^{\\frac{2-p}{2}}, R_2(x^{\\star}x)^{\\frac{2-p}{2}}`
        """
        pass


class FixedRankSpectralShiftRegularizationRule(RegularizationRule):
    def __init__(
        self,
        rank_estimate: int,
        minimal_shift=1e-9,
        initial_shift_parameter=1.0,
        svd_engine: SVDEngine = ScipySVDEngine,
    ):
        self._minimal_shift = minimal_shift
        self._shift_parameter = initial_shift_parameter
        self._svd_engine = svd_engine
        self._rank_estimate = rank_estimate

    @property
    def rank_estimate(self):
        return self._rank_estimate

    @property
    def shift_parameter(self):
        return self._shift_parameter

    def compute_regularized_inverse_weights(
        self, x: np.ndarray, schatten_p_parameter: np.float64
    ):
        """
        Regularization via spectral shift

        .. math::

            R_1 : \\mathbb{C}^{d_1 \\times d_1} \\rightarrow \\mathbb{C}^{d_1 \\times d_1}, x \\mapsto xx^{\\star}+\\varepsilon^2 \cdot Id_{d_1},

            R_2 : \\mathbb{C}^{d_2 \\times d_2} \\rightarrow \\mathbb{C}^{d_2 \\times d_2}, x \\mapsto x^{\\star}x +\\varepsilon^2 \cdot Id_{d_2}.

        Spectral shift :math:`\\varepsilon` is given by

        .. math::

            \\varepsilon = \\min(\\max(\\sigma_{\\text{rank_estimate}}(x), \\text{minimal_shift}), \\text{shift_parameter}),

        where :math:`\\sigma_{\\text{rank_estimate}}(x)` is the rank_estimate-th singular value of x.

        :param x: array of shape :math:`(d_1, d_2)`
        :param schatten_p_parameter:
        :return: :math:`\\left(xx^{\\star}+\\varepsilon \cdot Id_{d_1}\\right)^{\\frac{2-p}{2}}, \\left(x^{\\star}x +\\varepsilon \cdot Id_{d_2}\\right)^{\\frac{2-p}{2}}`
        """

        u, s, v = self._svd_engine.svd(x)
        min_non_zero_dim = min(u.shape[0], v.shape[0])
        regularized_singular_values = self._compute_regularized_singular_values(
            s, self._shift_parameter, schatten_p_parameter
        )
        left_inverse_weight = self._construct_one_sided_inverse_weight(
            u, regularized_singular_values, min_non_zero_dim
        )
        right_inverse_weight = self._construct_one_sided_inverse_weight(
            v, regularized_singular_values, min_non_zero_dim
        )
        self._update_shift_parameter(s)
        return left_inverse_weight, right_inverse_weight

    def _update_shift_parameter(self, singular_values):
        self._shift_parameter = min(
            max(singular_values[self.rank_estimate], self._minimal_shift),
            self._shift_parameter,
        )

    def _construct_one_sided_inverse_weight(
        self,
        singular_vectors: np.ndarray,
        regularized_singular_values,
        min_non_zero_dim,
    ):
        dimension = singular_vectors.shape[0]
        diagonal_matrix = self._construct_shifted_diagonal_matrix(
            dimension, regularized_singular_values, min_non_zero_dim
        )
        return singular_vectors @ diagonal_matrix @ singular_vectors.conj().transpose()

    @staticmethod
    def _construct_shifted_diagonal_matrix(
        dimension, regularized_singular_values, min_non_zero_dim
    ):
        _regularized_singular_values = np.zeros(dimension)
        _regularized_singular_values[:min_non_zero_dim] = regularized_singular_values
        _regularized_singular_values[min_non_zero_dim:] = 0
        return diags(_regularized_singular_values, shape=(dimension, dimension))

    @staticmethod
    def _compute_regularized_singular_values(
        singular_values, regularization_parameter, schatten_p_parameter
    ):
        return np.power(
            np.power(singular_values, 2) + regularization_parameter ** 2,
            (2 - schatten_p_parameter) / 2.0,
        )



from enum import Enum
from typing import Tuple, List, Union, Callable, Optional
import numpy as np
from scipy.sparse import csr_matrix, kron, eye
from scipy.sparse.linalg import LinearOperator, aslinearoperator


class MatrixOperatorCompatibility:
    """
    Compatibility check for objects of :class:`~MatrixOperator`
    """

    @staticmethod
    def compatible_for_addition(
        first_operator: "MatrixOperator", second_operator: "MatrixOperator"
    ):

        same_input_shape = first_operator.input_shape == second_operator.input_shape
        same_flattened_shape = (
            first_operator.flattened_operator.shape
            == second_operator.flattened_operator.shape
        )
        same_dtype = MatrixOperatorCompatibility._same_dtype(
            first_operator, second_operator
        )
        same_indexing_order = MatrixOperatorCompatibility._same_indexing_order(
            first_operator, second_operator
        )
        return (
            same_flattened_shape
            and same_input_shape
            and same_dtype
            and same_indexing_order
        )

    @staticmethod
    def compatible_for_composition(
        first_operator: "MatrixOperator", second_operator: "MatrixOperator"
    ):
        same_dtype = MatrixOperatorCompatibility._same_dtype(
            first_operator, second_operator
        )
        same_indexing_order = MatrixOperatorCompatibility._same_indexing_order(
            first_operator, second_operator
        )
        compatible_shapes = first_operator.input_shape == second_operator.output_shape
        return same_dtype and same_indexing_order and compatible_shapes

    @staticmethod
    def _same_dtype(
        first_operator: "MatrixOperator", second_operator: "MatrixOperator"
    ):
        return (
            first_operator.flattened_operator.dtype
            == second_operator.flattened_operator.dtype
        )

    @staticmethod
    def _same_indexing_order(
        first_operator: "MatrixOperator", second_operator: "MatrixOperator"
    ):
        return first_operator.order == second_operator.order


class MatrixOperator:
    class IndexingOrder(Enum):
        """
        Enum for order argument for :func:`numpy.reshape` and :func:`numpy.flatten`
        """
        ROW_MAJOR = "C"
        COLUMN_MAJOR = "F"

    def __init__(
        self,
        flattened_operator: Union[LinearOperator, np.ndarray, csr_matrix],
        input_shape: Tuple[int, int],
        output_shape: Tuple[int, int],
        order: IndexingOrder,
        representing_matrix: Optional[Union[np.ndarray, csr_matrix]] = None,
    ):
        """
        Represents an operator

        .. math::

            \\Phi: \\mathbb{C}^{d_1 \\times d_2} \\rightarrow \\mathbb{C}^{m_1 \\times m_2}.

        Takes care on flattening and reshaping. Wraps a scipy linear operator or numpy array.

        :param flattened_operator: Represents the flattened map

            .. math::

                \\varphi: \\mathbb{C}^{d_1 \\cdot d_2} \\rightarrow \\mathbb{C}^{m_1 \\cdot m_2},

            such that :math:`\\varphi(\\operatorname{vec}(x)) = \\operatorname{vec}(\\Phi(x))`.

        :param input_shape: :math:`(d_1,d_2)`
        :param output_shape: :math:`(m_1,m_2)`
        :param order: order for vectorization :math:`\\operatorname{vec}`
        :param representing_matrix: matrix representation of flattened operator

        """
        self._output_shape = output_shape
        self._input_shape = input_shape
        self._order = order
        self._representing_matrix = representing_matrix
        if isinstance(flattened_operator, np.ndarray):
            self._representing_matrix = flattened_operator
            flattened_operator = aslinearoperator(flattened_operator)
        self._flattened_operator = flattened_operator

    def __add__(self, other: "MatrixOperator"):
        if isinstance(other, MatrixOperator):
            if MatrixOperatorCompatibility.compatible_for_addition(self, other):
                representing_matrix = None
                if (
                    self._representing_matrix is not None
                    and other._representing_matrix is not None
                ):
                    representing_matrix = (
                        self._representing_matrix + other._representing_matrix
                    )
                return MatrixOperator(
                    self._flattened_operator + other._flattened_operator,
                    self._input_shape,
                    self._output_shape,
                    self._order,
                    representing_matrix,
                )
            raise ValueError(f"Operators are not compatible")
        return NotImplemented

    def __call__(self, x):
        return self * x

    def __mul__(self, x):
        return self.dot(x)

    def dot(self, x):
        """Matrix-matrix multiplication (operator application) or operator composition.

        Parameters
        ----------
        x : array_like of shape :math:`(d_1, d_2)` or MatrixOperator compatible for composition or scalar

        Returns
        -------
            :math:`\\Phi(x)` : array or MatrixOperator that represents
            the result of applying this linear operator on x.

        """
        if isinstance(x, MatrixOperator):
            if MatrixOperatorCompatibility.compatible_for_composition(self, x):
                representing_matrix = None
                if (
                    self.representing_matrix is not None
                    and x.representing_matrix is not None
                ):
                    representing_matrix = (
                        self.representing_matrix @ x.representing_matrix
                    )
                return MatrixOperator(
                    self.flattened_operator.dot(x.flattened_operator),
                    x.input_shape,
                    self.output_shape,
                    self.order,
                    representing_matrix,
                )
        elif np.isscalar(x):
            return MatrixOperator._from_callable(self, lambda op: x * op)
        else:
            if x.shape == self.input_shape:
                x_flattened = x.flatten(order=self.order.value)
                return self.flattened_operator(x_flattened).reshape(
                    self.output_shape, order=self.order.value
                )
            else:
                raise ValueError(
                    f"expected 2-d array or matrix of shape{self.input_shape}, got {x}"
                )

    def __matmul__(self, other):
        if np.isscalar(other):
            raise ValueError("Scalar operands are not allowed, " "use '*' instead")
        return self.__mul__(other)

    def __rmatmul__(self, other):
        if np.isscalar(other):
            raise ValueError("Scalar operands are not allowed, " "use '*' instead")
        return self.__rmul__(other)

    def __rmul__(self, x):
        if np.isscalar(x):
            return self.dot(x)
        else:
            return NotImplemented

    def __pow__(self, p):
        if np.isscalar(p):
            if self.input_shape != self.output_shape:
                raise ValueError("square LinearOperator expected, got %r" % self)
            return MatrixOperator._from_callable(self, lambda op: op ** p)
        else:
            return NotImplemented

    def __neg__(self):
        return MatrixOperator._from_callable(self, lambda op: -op)

    def __sub__(self, x):
        return self.__add__(-x)

    def __repr__(self):
        if self.flattened_operator.dtype is None:
            dt = "unspecified dtype"
        else:
            dt = "dtype=" + str(self.flattened_operator.dtype)

        return f"{(self.output_shape, self.input_shape, self.__class__.__name__, dt)}"

    @classmethod
    def _from_callable(
        cls,
        matrix_operator: "MatrixOperator",
        fcn: Callable[
            [Union[LinearOperator, np.ndarray]],
            Union[LinearOperator, np.ndarray],
        ],
    ):
        """
        Instantiation of a :class: ~`MatrixOperator` from an :class: ~`MatrixOperator` object and a callable
        :param matrix_operator:
        :param fcn:
        :return:
        """
        representing_matrix = None
        if matrix_operator.representing_matrix is not None:
            representing_matrix = fcn(matrix_operator.representing_matrix)
        return cls(fcn(matrix_operator.flattened_operator),
                   matrix_operator.input_shape,
                   matrix_operator.output_shape,
                   matrix_operator.order,
                   representing_matrix,
                   )

    @property
    def shape(self):
        return self._input_shape, self._output_shape

    @property
    def output_shape(self):
        return self._output_shape

    @property
    def input_shape(self):
        return self._input_shape

    @property
    def order(self):
        return self._order

    @property
    def flattened_operator(self):
        return self._flattened_operator

    @property
    def representing_matrix(self):
        return self._representing_matrix

    def adjoint(self):
        """Hermitian adjoint.

        Returns a :class:`~MatrixOperator` that represents the Hermitian adjoint of this one

        Can be abbreviated self.H instead of self.adjoint().

        Returns
        -------
        :math:`\\Phi^{\\star}` : :class:`~MatrixOperator`
            Hermitian adjoint of self.
        """
        return self._adjoint()

    H = property(adjoint)

    def transpose(self):
        """Transpose this linear operator.

        Returns a :class:`~MatrixOperator` that represents the transpose of this one.
        Can be abbreviated self.T instead of self.transpose().

        Returns
        -------
        :math:`\\Phi^{\\star}` : :class:`~MatrixOperator`
             adjoint of self.
        """

        return self._transpose()

    T = property(transpose)

    def _adjoint(self):
        """Default implementation of _adjoint"""
        representing_matrix = None
        if self.representing_matrix is not None:
            representing_matrix = self._representing_matrix.conjugate().transpose()
        return MatrixOperator(
            self._flattened_operator.H,
            self._output_shape,
            self._input_shape,
            self.order,
            representing_matrix,
        )

    def _transpose(self):
        """ Default implementation of _transpose"""
        return MatrixOperator._from_callable(self, lambda op: op.T)


class SamplingOperator(LinearOperator):
    def __init__(self, indices: Union[List[int], np.ndarray], input_dimension: int):
        """

        :type input_dimension:
        :type indices: linear indices
        """
        super().__init__(dtype=np.float64, shape=(len(indices), input_dimension))
        self._indices = indices
        self._sampling_matrix = self._construct_sampling_matrix()
        self._input_dimension = input_dimension

    @property
    def indices(self):
        return self._indices

    @property
    def input_dimension(self):
        return self._input_dimension

    @property
    def sampling_matrix(self):
        return self._sampling_matrix

    def _matvec(self, x: np.array):
        return x[self.indices]

    def _rmatvec(self, x: np.array):
        return csr_matrix(
            (x.flatten(), (self.indices, np.zeros(len(self.indices)))),
            shape=(self.shape[1], 1),
            dtype=self.dtype,
        ).toarray()

    def _construct_sampling_matrix(self):
        data = np.ones(len(self.indices))
        row_indices = list(range(len(self.indices)))
        column_indices = self.indices
        return csr_matrix((data, (row_indices, column_indices)), shape=self.shape)

    @classmethod
    def from_matrix_indices(
        cls, row_indices: List[int], column_indices: List[int], shape: Tuple[int, int],
            order: MatrixOperator.IndexingOrder = MatrixOperator.IndexingOrder.COLUMN_MAJOR
    ):
        if len(row_indices) != len(column_indices):
            raise ValueError("Row and column indices must be of same length")

        return cls(
            np.ravel_multi_index((row_indices, column_indices), shape, order=order.value),
            input_dimension=np.product(shape),
        )


class SamplingMatrixOperator(MatrixOperator):
    def __init__(
        self,
        row_indices: List[int],
        column_indices: List[int],
        shape: Tuple[int, int],
        order: MatrixOperator.IndexingOrder = MatrixOperator.IndexingOrder.COLUMN_MAJOR,
    ):
        flattened_operator = SamplingOperator.from_matrix_indices(
            row_indices, column_indices, shape, order
        )
        super().__init__(
            flattened_operator,
            shape,
            (len(row_indices), 1),
            order,
            flattened_operator.sampling_matrix,
        )

    @classmethod
    def from_index_tuples(cls, indices: List[int], shape: Tuple[int, int]):
        row_indices = []
        column_indices = []
        for (row_idx, col_idx) in indices:
            row_indices.append(row_idx)
            column_indices.append(col_idx)
        return cls(row_indices, column_indices, shape)


class InverseWeightOperator(MatrixOperator):
    def __init__(
        self,
        left_inverse_weight: np.ndarray,
        right_inverse_weight: np.ndarray,
        order,
    ):
        """
        Constructs inverse weight operator from one-sided inverse weights.

        :param left_inverse_weight:
        :param right_inverse_weight:
        :param order:
        """
        shape = (left_inverse_weight.shape[0], right_inverse_weight.shape[0])
        weight_matrix = 0.5 * (
            kron(eye(right_inverse_weight.shape[0]), left_inverse_weight)
            + kron(right_inverse_weight, eye(left_inverse_weight.shape[0]))
        )
        flattened_operator = aslinearoperator(weight_matrix)
        super().__init__(
            flattened_operator,
            shape,
            shape,
            order,
            weight_matrix,

        )



from typing import Optional
import time


class StopWatch:
    """
    A simple stop watch to be used as a context manager
    """
    def __init__(self):
        self._start: Optional[float] = None
        self._stop: Optional[float] = None

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def duration(self):
        return self._stop - self._start

    def __enter__(self):
        self._start = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stop = time.time()


from abc import ABC, abstractmethod
from typing import Union
import numpy as np
from scipy.sparse.linalg import LinearOperator as ScipyLinearOperator, cg, lsqr, spsolve, minres
from scipy.sparse import eye, csr_matrix
from .operators import MatrixOperator


class WeightedLeastSquaresSolver(ABC):
    """
    Abstract base class for weighted least square solvers
    """

    def solve(
        self,
        operator: Union[np.ndarray, MatrixOperator],
        inverse_weight_operator: MatrixOperator,
        data: np.ndarray,
    ):
        """

        :param operator: :class:`~MatrixOperator`
        :param inverse_weight_operator: :class:`~MatrixOperator`
        :param data:
        :return:
        """
        combined_operator = operator * inverse_weight_operator * operator.H
        x = self._solve_linear_equation(combined_operator.flattened_operator, data)
        x = x.reshape(
            operator.output_shape,
            order=operator.order.value,
        )
        return (inverse_weight_operator * operator.H)(x)

    @abstractmethod
    def _solve_linear_equation(
        self, operator: MatrixOperator, data: np.ndarray
    ) -> np.ndarray:
        pass


class ScipyCgWeightedLeastSquaresSolver(WeightedLeastSquaresSolver):
    def __init__(self, tol=1e-10):
        self._tol = tol

    def _solve_linear_equation(
        self, operator: ScipyLinearOperator, data: np.ndarray
    ) -> np.ndarray:
        x, _ = cg(operator, data, tol=self._tol)
        return x


class ScipyLsqrWeightedLeastSquareSolver(WeightedLeastSquaresSolver):

    def _solve_linear_equation(self, operator: MatrixOperator, data: np.ndarray) -> np.ndarray:
        x = lsqr(operator, data)
        return x[0]


class ScipyDirect(WeightedLeastSquaresSolver):

    def _solve_linear_equation(self, operator: MatrixOperator, data: np.ndarray) -> np.ndarray:
        x = spsolve(csr_matrix(operator @ np.eye(*operator.shape)), data)
        return x


class ScipyMinRes(WeightedLeastSquaresSolver):
    def __init__(self, tol: float = 1e-5):
        self._tol = tol

    def _solve_linear_equation(self, operator: MatrixOperator, data: np.ndarray) -> np.ndarray:
        x = minres(operator, data, tol=self._tol)
        return x[0]


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['hmirls']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.24.1,<2.0.0', 'scipy>=1.10.0,<2.0.0']

setup_kwargs = {
    'name': 'hmirls',
    'version': '0.1.0',
    'description': 'Library for Schatten-p norm minimization via iteratively reweighted least squares',
    'long_description': '# Harmonic Mean Iteratively Reweighted Least Squares for Low-Rank Matrix Recovery\n\nThis repository contains python code to implement a basic variant of the Harmonic Mean Iteratively Reweighted Least Squares (HM-IRLS) algorithm for low-rank matrix recovery, in particular for the low-rank matrix completion problem, described in the paper:\n\n> C. Kümmerle, J. Sigl.\n> "Harmonic Mean Iteratively Reweighted Least Squares for Low-Rank Matrix Recovery", Journal of Machine Learning Research (JMLR) volume 19, number 47, pages 1-49, 2018.\n> Available online: https://jmlr.org/papers/volume19/17-244/17-244.pdf\n\n## Version history\n* Version 0.0.1, 10/25/2020\n\n## Author\nKristof Schröder\n\n## Documentation\n> https://hmirls.readthedocs.io/en/latest/',
    'author': 'Kristof Schroeder',
    'author_email': 'kristof_schroeder@web.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np
import scipy


class SVDEngine(ABC):
    @staticmethod
    @abstractmethod
    def svd(x: np.ndarray, *args, **kwargs) -> Tuple[np.ndarray, np.array, np.ndarray]:
        pass


class ScipySVDEngine(SVDEngine):
    @staticmethod
    def svd(x: np.ndarray, *args, **kwargs) -> Tuple[np.ndarray, np.array, np.ndarray]:
        left_sing_vec, sing_val, right_sing_vec_herm = scipy.linalg.svd(
            x, full_matrices=True
        )
        return left_sing_vec, sing_val, right_sing_vec_herm.conj().transpose()


class NumpySVDEngine(SVDEngine):
    @staticmethod
    def svd(x: np.ndarray, *args, **kwargs) -> Tuple[np.ndarray, np.array, np.ndarray]:
        pass
